@@ -8,7 +8,7 @@ import os
 import re
 import time
 import urllib.parse
-from typing import TYPE_CHECKING, Any, Dict, Mapping, Optional, Sequence, Tuple
+from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Sequence, Tuple
 
 import xmltodict
 
@@ -24,7 +24,7 @@ OAUTH_TOKEN = "oauth_token"
 SHARED_KEY = "shared_key"
 
 
-def load_credentials() -> Dict[str, str]:
+def load_credentials() -> Dict[str, Any]:
     # AZURE_STORAGE_KEY seems to be the environment variable mentioned by the az cli
     # AZURE_STORAGE_ACCOUNT_KEY is mentioned elsewhere on the internet
     for varname in ["AZURE_STORAGE_KEY", "AZURE_STORAGE_ACCOUNT_KEY"]:
@@ -66,14 +66,6 @@ def load_credentials() -> Dict[str, str]:
     # https://mikhail.io/2019/07/how-azure-cli-manages-access-tokens/
     default_creds_path = os.path.expanduser("~/.azure/accessTokens.json")
     if os.path.exists(default_creds_path):
-        default_profile_path = os.path.expanduser("~/.azure/azureProfile.json")
-        if not os.path.exists(default_profile_path):
-            raise RuntimeError(f"Missing default profile path: '{default_profile_path}'")
-        with open(default_profile_path, "rb") as g:
-            # this file has a UTF-8 BOM
-            profile = json.loads(g.read().decode("utf-8-sig"))
-            subscriptions = [sub["id"] for sub in profile["subscriptions"]]
-
         with open(default_creds_path) as f:
             tokens = json.load(f)
             best_token = None
@@ -81,12 +73,11 @@ def load_credentials() -> Dict[str, str]:
                 if best_token is None:
                     best_token = token
                 else:
-                    if token["expiresOn"] > best_token["expiresOn"]:
+                    # expiresOn may be missing for tokens from service principals
+                    if token.get("expiresOn", "") > best_token.get("expiresOn", ""):
                         best_token = token
             if best_token is not None:
-                token = best_token.copy()
-                token["subscriptions"] = subscriptions
-                return token
+                return best_token
 
     raise RuntimeError(
         """Azure credentials not found, please do one of the following:
@@ -104,22 +95,45 @@ def load_credentials() -> Dict[str, str]:
     )
 
 
-async def get_access_token(account: str) -> Tuple[Any, float]:
+def load_stored_subscription_ids() -> List[str]:
+    """Return a list of subscription ids from the local azure profile.
+
+    The default subscription will appear first in the list.
+
+    """
+    default_profile_path = os.path.expanduser("~/.azure/azureProfile.json")
+    if not os.path.exists(default_profile_path):
+        return []
+
+    with open(default_profile_path, "rb") as f:
+        # this file has a UTF-8 BOM
+        profile = json.loads(f.read().decode("utf-8-sig"))
+
+    subscriptions = profile["subscriptions"]
+    subscriptions.sort(key=lambda x: x["isDefault"], reverse=True)
+    return [sub["id"] for sub in subscriptions]
+
+
+async def get_access_token(cache_key: Tuple[str, Optional[str]]) -> Tuple[Any, float]:
+    account, container = cache_key
+
     now = time.time()
     creds = load_credentials()
+
     if "storageAccountKey" in creds:
         if "account" in creds:
             if creds["account"] != account:
                 raise RuntimeError(
-                    f"Found credentials for account '{creds['account']}' but needed credentials "
-                    f"for account '{account}'"
+                    f"Provided storage account key for account '{creds['account']}' via "
+                    f"environment variables, but needed credentials for account '{account}'"
                 )
         auth = (SHARED_KEY, creds["storageAccountKey"])
-        if await can_access_account(account, auth):
+        if await can_access_account(account, container, auth):
             return (auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)
         raise RuntimeError(
             f"Found storage account key, but it was unable to access storage account: '{account}'"
         )
+
     if "refreshToken" in creds:
         # we have a refresh token, convert it into an access token for this account
         req = create_access_token_request(
@@ -148,57 +162,64 @@ async def get_access_token(account: str) -> Tuple[Any, float]:
         auth = (OAUTH_TOKEN, result["access_token"])
 
         # for some azure accounts this access token does not work, check if it works
-        if await can_access_account(account, auth):
+        if await can_access_account(account, container, auth):
             return (auth, now + float(result["expires_in"]))
 
         # it didn't work, fall back to getting the storage keys
-        storage_account_key_auth = await get_storage_account_key(account=account, creds=creds)
+        storage_account_key_auth = await get_storage_account_key(
+            account=account, creds=creds, container_hint=container
+        )
         if storage_account_key_auth is not None:
             return (storage_account_key_auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)
 
-        raise RuntimeError(
-            f"Could not find any credentials that grant access to storage account: '{account}'"
+    if "appId" in creds:
+        # we have a service principal, get an oauth token
+        req = create_access_token_request(creds=creds, scope="https://storage.azure.com/")
+
+        async with req.execute() as resp:
+            result = await resp.json()
+        auth = (OAUTH_TOKEN, result["access_token"])
+        if await can_access_account(account, container, auth):
+            return (auth, now + float(result["expires_in"]))
+
+        # it didn't work, fall back to getting the storage keys
+        storage_account_key_auth = await get_storage_account_key(
+            account=account, creds=creds, container_hint=container
         )
+        if storage_account_key_auth is not None:
+            return (storage_account_key_auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)
 
-    # we have a service principal, get an oauth token
-    req = create_access_token_request(creds=creds, scope="https://storage.azure.com/")
-
-    async with req.execute() as resp:
-        result = await resp.json()
-    auth = (OAUTH_TOKEN, result["access_token"])
-    if await can_access_account(account, auth):
-        return (auth, now + float(result["expires_in"]))
-
-    storage_account_key_auth = await get_storage_account_key(account=account, creds=creds)
-    if storage_account_key_auth is not None:
-        return (storage_account_key_auth, now + AZURE_SHARED_KEY_EXPIRATION_SECONDS)
     raise RuntimeError(
         f"Could not find any credentials that grant access to storage account: '{account}'"
     )
 
 
-async def can_access_account(account: str, auth: Tuple[str, str]) -> bool:
+async def can_access_account(account: str, container: Optional[str], auth: Tuple[str, str]) -> bool:
     from .request import Request, azurify_request
 
-    req = Request(
-        method="GET",
-        url=f"https://{account}.blob.core.windows.net",
-        params={"comp": "list", "maxresults": "1"},
-        success_codes=(200, 403),
-    )
-    req = await azurify_request(req, auth=auth)
+    if not container:
+        # if a container isn't specified, check that we can list the storage account
+        req = Request(
+            method="GET",
+            url=f"https://{account}.blob.core.windows.net",
+            params={"comp": "list", "maxresults": "1"},
+            success_codes=(200, 403),
+        )
+        req = await azurify_request(req, auth=auth)
 
-    async with req.execute() as resp:
-        if resp.status == 403:
-            return False
-        data = await resp.read()
+        async with req.execute() as resp:
+            if resp.status == 403:
+                return False
+            data = await resp.read()
 
-    out = xmltodict.parse(data)
-    if out["EnumerationResults"]["Containers"] is None:
-        # there are no containers in this storage account
-        # we can't test if we can access this storage account or not, so presume we can
-        return True
-    container = out["EnumerationResults"]["Containers"]["Container"]["Name"]
+        out = xmltodict.parse(data)
+        if out["EnumerationResults"]["Containers"] is None:
+            # there are no containers in this storage account
+            # we can't test if we can access this storage account or not, so presume we can
+            return True
+
+        # then also test that we can list a container. this is perhaps unnecessary...
+        container = out["EnumerationResults"]["Containers"]["Container"]["Name"]
 
     # https://myaccount.blob.core.windows.net/mycontainer?restype=container&comp=list
     req = Request(
@@ -250,8 +271,61 @@ def create_access_token_request(
     )
 
 
+async def get_storage_account_id_with_subscription(
+    subscription_id: str, account: str, auth: Tuple[str, str]
+) -> Optional[str]:
+    from .request import Request, azurify_request
+
+    req = Request(
+        method="GET",
+        url=f"https://management.azure.com/subscriptions/{subscription_id}/providers/Microsoft.Storage/storageAccounts",
+        params={"api-version": "2019-04-01"},
+        success_codes=(200, 401, 403),
+    )
+    req = await azurify_request(req, auth=auth)
+
+    async with req.execute() as resp:
+        if resp.status in (401, 403):
+            # we aren't allowed to query this for this subscription, skip it
+            return None
+        out = await resp.json()
+
+    # search for the storage account
+    return next((obj["id"] for obj in out["value"] if obj["name"] == account), None)
+
+
+async def get_storage_account_id(account: str, auth: Tuple[str, str]) -> Optional[str]:
+    from .request import Request, azurify_request
+
+    stored_subscription_ids = load_stored_subscription_ids()
+    for subscription_id in stored_subscription_ids:
+        storage_account_id = await get_storage_account_id_with_subscription(
+            subscription_id, account, auth
+        )
+        if storage_account_id:
+            return storage_account_id
+
+    req = Request(
+        method="GET",
+        url="https://management.azure.com/subscriptions",
+        params={"api-version": "2020-01-01"},
+    )
+    req = await azurify_request(req, auth=auth)
+    async with req.execute() as resp:
+        result = await resp.json()
+    subscription_ids = {item["subscriptionId"] for item in result["value"]}
+
+    for subscription_id in subscription_ids - set(stored_subscription_ids):
+        storage_account_id = await get_storage_account_id_with_subscription(
+            subscription_id, account, auth
+        )
+        if storage_account_id:
+            return storage_account_id
+    return None
+
+
 async def get_storage_account_key(
-    account: str, creds: Mapping[str, str]
+    account: str, creds: Mapping[str, Any], container_hint: Optional[str] = None
 ) -> Optional[Tuple[Any, float]]:
     from .request import Request, azurify_request
 
@@ -262,68 +336,39 @@ async def get_storage_account_key(
         result = await resp.json()
     auth = (OAUTH_TOKEN, result["access_token"])
 
-    # get a list of subscriptions so we can query each one for storage accounts
+    storage_account_id = await get_storage_account_id(account, auth)
+    if not storage_account_id:
+        return None
+
     req = Request(
-        method="GET",
-        url="https://management.azure.com/subscriptions",
-        params={"api-version": "2020-01-01"},
+        method="POST",
+        url=f"https://management.azure.com{storage_account_id}/listKeys",
+        params={"api-version": "2019-04-01"},
     )
     req = await azurify_request(req, auth=auth)
+
     async with req.execute() as resp:
         result = await resp.json()
-    subscription_ids = [item["subscriptionId"] for item in result["value"]]
-
-    for subscription_id in subscription_ids:
-        # get a list of storage accounts
-        req = Request(
-            method="GET",
-            url=f"https://management.azure.com/subscriptions/{subscription_id}/providers/Microsoft.Storage/storageAccounts",
-            params={"api-version": "2019-04-01"},
-            success_codes=(200, 401, 403),
-        )
-        req = await azurify_request(req, auth=auth)
-
-        async with req.execute() as resp:
-            if resp.status in (401, 403):
-                # we aren't allowed to query this for this subscription, skip it
-                # it's unclear if this is still necessary since we query for subscriptions first
-                continue
-            out = await resp.json()
-
-        # check if we found the storage account we are looking for
-        for obj in out["value"]:
-            if obj["name"] == account:
-                storage_account_id = obj["id"]
-                break
-        else:
-            continue
-
-        req = Request(
-            method="POST",
-            url=f"https://management.azure.com{storage_account_id}/listKeys",
-            params={"api-version": "2019-04-01"},
-        )
-        req = await azurify_request(req, auth=auth)
-
-        async with req.execute() as resp:
-            result = await resp.json()
-        for key in result["keys"]:
-            if key["permissions"] == "FULL":
-                storage_key_auth = (SHARED_KEY, key["value"])
-                if await can_access_account(account, storage_key_auth):
-                    return storage_key_auth
-                raise RuntimeError(
-                    f"Found storage account key, but it was unable to access storage account: '{account}'"
-                )
-        raise RuntimeError(
-            f"Storage account was found, but storage account keys were missing: '{account}'"
-        )
-    return None
+    for key in result["keys"]:
+        if key["permissions"] == "FULL":
+            storage_key_auth = (SHARED_KEY, key["value"])
+            if await can_access_account(account, container_hint, storage_key_auth):
+                return storage_key_auth
+            raise RuntimeError(
+                f"Found storage account key, but it was unable to access storage account: '{account}'"
+            )
+    raise RuntimeError(
+        f"Storage account was found, but storage account keys were missing: '{account}'"
+    )
 
 
 def sign_request_with_shared_key(request: Request, key: str) -> str:
     # https://docs.microsoft.com/en-us/rest/api/storageservices/authorize-with-shared-key
-    parsed_url = urllib.parse.urlparse(request.url)
+    from yarl import URL
+
+    # Figuring out that shared key authorisation sometimes breaks because aiohttp is a little over
+    # clever about canonicalising URLs was not fun; preemptively canonicalise
+    parsed_url = urllib.parse.urlparse(str(URL(request.url)))
     storage_account = parsed_url.netloc.split(".")[0]
     blob = parsed_url.path[1:]
 
@@ -371,39 +416,51 @@ def sign_request_with_shared_key(request: Request, key: str) -> str:
     return f"SharedKey {storage_account}:{signature}"
 
 
-async def get_sas_token(account: str) -> Tuple[Any, float]:
+async def get_sas_token(cache_key: Tuple[str, Optional[str]]) -> Tuple[Any, float]:
     from .globals import config
-    from .request import azurify_request
+    from .request import Request, azurify_request
 
-    auth = await config.azure_access_token_manager.get_token(key=account)
+    auth = await config.azure_access_token_manager.get_token(key=cache_key)
+    account, container = cache_key
 
     if auth[0] != OAUTH_TOKEN:
-        raise RuntimeError("Only oauth tokens can be used to get SAS tokens")
+        cmd = (
+            f"az storage container list --auth-mode login --account-name {account}"
+            if container is None
+            else f"az storage blob list --auth-mode login --account-name {account} --container {container}"
+        )
+        raise RuntimeError(
+            "Only OAuth tokens can be used to get SAS tokens. You should set the "
+            "Blob Data Reader or Storage Blob Data Contributor IAM role. You can run "
+            f"`{cmd}` to confirm that the missing role is the issue."
+        )
 
-    req = create_user_delegation_sas_request(account=account)
-    req = await azurify_request(req, auth=auth)
-
-    async with req.execute() as resp:
-        data = await resp.read()
-
-    out = xmltodict.parse(data)
-    return out["UserDelegationKey"], time.time() + AZURE_SAS_TOKEN_EXPIRATION_SECONDS
-
-
-def create_user_delegation_sas_request(account: str) -> Request:
     # https://docs.microsoft.com/en-us/rest/api/storageservices/create-user-delegation-sas
-    from .request import Request
-
     now = datetime.datetime.utcnow()
     start = (now + datetime.timedelta(hours=-1)).strftime("%Y-%m-%dT%H:%M:%SZ")
     expiration = now + datetime.timedelta(days=6)
     expiry = expiration.strftime("%Y-%m-%dT%H:%M:%SZ")
-    return Request(
-        url=f"https://{account}.blob.core.windows.net/",
-        method="POST",
-        params=dict(restype="service", comp="userdelegationkey"),
-        data={"KeyInfo": {"Start": start, "Expiry": expiry}},
+    req = await azurify_request(
+        Request(
+            url=f"https://{account}.blob.core.windows.net/",
+            method="POST",
+            params=dict(restype="service", comp="userdelegationkey"),
+            data={"KeyInfo": {"Start": start, "Expiry": expiry}},
+            success_codes=(200, 403),
+        ),
+        auth=auth,
     )
+    async with req.execute() as resp:
+        if resp.status == 403:
+            raise RuntimeError(
+                f"You do not have permission to generate an SAS token for account {account}. "
+                "Try setting the Storage Blob Delegator or Storage Blob Data Contributor IAM role "
+                "at the account level."
+            )
+        data = await resp.read()
+
+    out = xmltodict.parse(data)
+    return out["UserDelegationKey"], time.time() + AZURE_SAS_TOKEN_EXPIRATION_SECONDS
 
 
 async def generate_signed_url(path: AzurePath) -> Tuple[str, datetime.datetime]:
@@ -412,7 +469,7 @@ async def generate_signed_url(path: AzurePath) -> Tuple[str, datetime.datetime]:
     # https://docs.microsoft.com/en-us/rest/api/storageservices/service-sas-examples
     from .globals import config
 
-    key = await config.azure_sas_token_manager.get_token(key=path.account)
+    key = await config.azure_sas_token_manager.get_token(key=(path.account, path.container or None))
     params = {
         "st": key["SignedStart"],
         "se": key["SignedExpiry"],
